from __future__ import annotations

import contextlib
import errno
import typing

import pr2modules.iproute.linux
import pr2modules.netlink.exceptions
from capport import cptypes


@contextlib.asynccontextmanager
async def connect():
    yield NeighborController()


# TODO: run blocking iproute calls in a different thread?
class NeighborController:
    def __init__(self):
        self.ip = pr2modules.iproute.linux.IPRoute()

    async def get_neighbor(
        self,
        address: cptypes.IPAddress,
        *,
        index: int=0,  # interface index
        flags: int=0,
    ) -> typing.Optional[pr2modules.iproute.linux.ndmsg.ndmsg]:
        if not index:
            route = await self.get_route(address)
            if route is None:
                return None
            index = route.get_attr(route.name2nla('oif'))
        try:
            return self.ip.neigh('get', dst=str(address), ifindex=index, state='none')[0]
        except pr2modules.netlink.exceptions.NetlinkError as e:
            if e.code == errno.ENOENT:
                return None
            raise

    async def get_neighbor_mac(
        self,
        address: cptypes.IPAddress,
        *,
        index: int=0,  # interface index
        flags: int=0,
    ) -> typing.Optional[cptypes.MacAddress]:
        neigh = await self.get_neighbor(address, index=index, flags=flags)
        if neigh is None:
            return None
        mac = neigh.get_attr(neigh.name2nla('lladdr'))
        return cptypes.MacAddress.parse(mac)

    async def get_route(
        self,
        address: cptypes.IPAddress,
    ) -> typing.Optional[pr2modules.iproute.linux.rtmsg]:
        try:
            return self.ip.route('get', dst=str(address))[0]
        except pr2modules.netlink.exceptions.NetlinkError as e:
            if e.code == errno.ENOENT:
                return None
            raise
